

set -ex



pip check || true
pytest -v -k "not (test_url_is_accessible or test_classes or test_feature_types or test_feature_types or test_is_valid_file or test_num_examples or test_smoke or test_str_smoke or test_transforms or test_forward_backward or test_jit_forward_backward or test_get_model or test_get_model_builder or test_get_model_weights or test_weights_copyable or test_weights_deserializable or test_list_models or test_get_weight or test_naming_conventions or test_schema_meta_validation or test_transforms_jit or test_no_warn or test_pretrained_pos or test_pretrained_kw or test_equivalent_behavior_weights or test_multi_params or test_default_callable or test_pretrained_deprecation or test_detection_model)" test/
exit 0
